/*
Copyright (c) 2010, Paul Furgale and Chi Hay Tong
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are 
met:

* Redistributions of source code must retain the above copyright notice, 
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright 
  notice, this list of conditions and the following disclaimer in the 
  documentation and/or other materials provided with the distribution.
* The names of its contributors may not be used to endorse or promote 
  products derived from this software without specific prior written 
  permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "GpuSurfDetectorInternal.hpp"
#include "gpu_utils.h"
#include "assert_macros.hpp"
#include "fasthessian.h"
#include "gpu_area.h"
#include "keypoint_interpolation.h"
#include "non_max_suppression.h"
#include "orientation.h"
#include "descriptors.h"
#include <boost/lexical_cast.hpp>
#include <fstream>

namespace asrl {

  GpuSurfDetectorInternal::GpuSurfDetectorInternal(GpuSurfConfiguration config) : 
    m_initialized(false),
    m_config(config)
  {
    int deviceCount;
    int device;
    cudaError_t err;
    cudaGetDeviceCount(&deviceCount);
    ASRL_ASSERT_GT(deviceCount,0,"There are no CUDA capable devices present");
    
	
    err = cudaGetDevice(&device);
    ASRL_ASSERT_EQ(err,cudaSuccess, "Unable to get the CUDA device: " << cudaGetErrorString(err));		
    //std::cout << "Found device " << device << std::endl;
    err = cudaGetDeviceProperties(&m_deviceProp,device);
    ASRL_ASSERT_EQ(err,cudaSuccess, "Unable to get the CUDA device properties: " << cudaGetErrorString(err));		

    // Some more checking...
    ASRL_ASSERT_GE(m_deviceProp.major,1,"Minimum compute capability 1.1 is necessary");
    ASRL_ASSERT_GE(m_deviceProp.minor,1,"Minimum compute capability 1.1 is necessary");

    m_maxmin.init(ASRL_SURF_MAX_CANDIDATES,false);
    m_maxmin.memset(0);
    m_threshold.init(1);
    m_threshold.h_get()[0] = config.threshold;
    m_threshold.pushToDevice();
    m_hessianPointers.init(config.nOctaves);

    // Set the appropriate device.
    // This doesn't work:
    //CUDA_SAFE_CALL(cudaSetDevice(config.deviceId));

  }

  GpuSurfDetectorInternal::~GpuSurfDetectorInternal()
  {

  }

  void GpuSurfDetectorInternal::buildIntegralImage(cv::Mat & image)
  {

    if(!m_initialized)
      {
	initDetector(image.cols,image.rows);
      }
    
    m_intProcessor->process(image, *m_intImg);
    texturize_integral_image(m_intImg->d_get());
    init_globals(m_intImg->width(), m_intImg->height(), m_octaves);
  }

  void GpuSurfDetectorInternal::saveHessianBuffers(std::string const & basename)
  {

    for(unsigned o = 0; o < m_octaves.size(); o++)
      {
	m_octaves[o].pullHessianFromDevice();
	float * hptr = m_octaves[o].h_hessian();
	int intervalSize = m_octaves[o].width() * m_octaves[o].height();
	for(int i = 0; i < m_octaves[o].intervals(); i++, hptr += intervalSize)
	  {
	    // A stringstream used to build the file names.
	    std::stringstream sout;
	    sout << basename << "-octave-" << o << "-interval-" << i << ".bin";
	    std::ofstream fout(sout.str().c_str(),std::ios::binary);
	    ASRL_ASSERT(fout.good(),"Unable to open file \"" << sout.str() << "\" for writing");
	    int size[2];
	    size[0] = m_octaves[o].width();
	    size[1] = m_octaves[o].height();
	    fout.write(reinterpret_cast<const char *>(&size[0]),2*sizeof(int));
	    fout.write(reinterpret_cast<const char *>(hptr),intervalSize*sizeof(float));
	  }
      }
  }

  void GpuSurfDetectorInternal::saveIntegralImage(std::string const & basename)
  {
    float * iimg = m_intImg->h_get();
    std::stringstream sout;
    sout << basename << "-iimg.bin";
    std::ofstream fout(sout.str().c_str(),std::ios::binary);
    ASRL_ASSERT(fout.good(),"Unable to open file \"" << sout.str() << "\" for writing");
    int size[2];
    size[0] = m_intImg->width();
    size[1] = m_intImg->height();
    fout.write(reinterpret_cast<const char *>(&size[0]),2*sizeof(int));
    fout.write(reinterpret_cast<const char *>(iimg),m_intImg->width()*m_intImg->height() * sizeof(float));

  }


  void GpuSurfDetectorInternal::detectKeypoints()
  {
    run_fast_hessian(m_octaves, m_config.detector_threads_x, m_config.detector_threads_y);

    /////////////////////////////////////////////////////
    ////               NON MAX SUPRESSION              //
    /////////////////////////////////////////////////////
    do_non_max_suppression(m_octaves, m_features,m_threshold, m_config.nonmax_threads_x, m_config.nonmax_threads_y);		

    m_features.featureCounterMem().pullFromDevice();
 

    ////////////////////////////////////////////////////
    ////               MAX/MIN INTERP                 //
    ////////////////////////////////////////////////////
    interpolate_keypoints(m_features, m_hessianPointers, m_threshold);


    m_features.featureCounterMem().pullFromDevice();
   

    // Untextureizing the integral image causes the orientation and
    // descriptors to fail...Where is the right place to call this?
    //fh_untexturizeIntegral();    

  }

  void GpuSurfDetectorInternal::findOrientation()
  {
    find_orientation(m_features.deviceFeatures(), m_features.ftCount());
  }

  void GpuSurfDetectorInternal::findOrientationFast()
  {
    find_orientation_fast(m_features.deviceFeatures(), m_features.ftCount());
  }

  void GpuSurfDetectorInternal::computeDescriptors()
  {
    compute_descriptors(m_features.deviceDescriptors(), m_features.deviceFeatures(), m_features.ftCount());
  }

  void GpuSurfDetectorInternal::computeUprightDescriptors()
  {
    ASRL_THROW("Not implemented");
  }

  void GpuSurfDetectorInternal::getKeypoints(std::vector<cv::KeyPoint> & outKeypoints)
  {
    m_features.getKeypoints(outKeypoints);
//     m_features.featuresMem().pullFromDevice();

//     // This has an implicit "Thread synchronize"
//     int ftcount = m_features.ftCount();
//     //std::cout << "Found " << ftcount << " features\n";

//     outKeypoints.resize(ftcount);
//     if(ftcount > 0)
//       {
// 	Keypoint * keypoints = m_features.hostFeatures();		
// 	unsigned k = 0;
// 	for(int i = 0; i < ftcount; i++) {
// 	  cv::KeyPoint & k = outKeypoints[i];
// 	  Keypoint & ft = keypoints[i];
			
// 	  k.pt.x = ft.x;
// 	  k.pt.y = ft.y;
			
// 	  k.size = ft.size;

// 	  k.angle = ft.angle;

// 	  k.response = ft.response;
// 	  k.octave = ft.octave;

// 	}
//       }
  }


  void GpuSurfDetectorInternal::setKeypoints(std::vector<cv::KeyPoint> const & inKeypoints)
  {
    m_features.setKeypoints(inKeypoints);
  }

  void GpuSurfDetectorInternal::getDescriptors(std::vector<float> & outDescriptors)
  {
    m_features.descriptorsMem().pullFromDevice();
    // This has an implicit "Thread synchronize"
    int ftcount = m_features.ftCount();

    // Resize the destination buffer.
    outDescriptors.resize(descriptorSize() * ftcount);
    // Copy the descriptors into the buffer. AFAIK, all known std::vector implementations use
    // contiguous memory.
    memcpy(&outDescriptors[0],m_features.hostDescriptors(), descriptorSize() * ftcount * sizeof(float));

  }

  int GpuSurfDetectorInternal::descriptorSize()
  {
    return ASRL_SURF_DESCRIPTOR_DIM;
  }


  void GpuSurfDetectorInternal::initDetector(int width, int height) {
    
    m_intProcessor.reset(new GpuIntegralImageProcessor(width, height));
    
    m_intImg.reset(new GpuIntegralImage(width, height));
    
    // initialize the fast hessian parameters.
    m_dxx_width = 1 + m_config.l1;
    m_dxx_height = m_config.l2;

    
    // Initialize the fast hessian memory.
    m_octaves.resize(m_config.nOctaves);
    
    m_hessianPointers.init(m_config.nOctaves);
    
    for(int o = 0; o < m_config.nOctaves; o++)
      {
	m_octaves[o] = GpuSurfOctave(width,height,m_config.l1,m_config.l2,m_config.l3, m_config.l4, m_config.edgeScale, m_config.initialScale, o, m_config.initialStep, m_config.nIntervals);
	m_hessianPointers.h_get()[o] = m_octaves[o].d_hessian();	
      }

    
    m_hessianPointers.pushToDevice();
    
    m_initialized = true;
  }

} // namespace asrl
